/*
Copyright 2007, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2007
*/

/**
 * @file	ScatterWeb.SHT11.internals.h
 * @ingroup	libSHT11
 * @brief	SHT11 library
 * 
 * @author	Bastian Blywis	<blywis@inf.fu-berlin.de>
 * @date	Aug 2007
 * @since		1.1
 * @version	$Revision$
 * 
 * Header file containing private declarations used by the SHT11 library.
 */

#ifndef SHT11_INT_H_
#define SHT11_INT_H_

#include <ScatterWeb.System.h>
#include "ScatterWeb.SHT11.h"

enum {
	noACK = 0x00,
	ACK = 0x01
};

typedef enum _SHT11_command {
	MEASURE_TEMP = 	0x03,	// 0000 0011
	MEASURE_HUMI =	0x05,	// 0000 0101
	STATUS_REG_W =		0x06,	// 0000 0110
	STATUS_REG_R =		0x07,	// 0000 0111
	RESET_SHT11 =			0x1E	// 0001 1110
} SHT11_command;

enum SHT11_statusFlags {
	MEASURE_RESOLUTION = 	0x01,
	OTP_RELOAD =						0x02,
	HEATER =								0x04,
	LOW_VOLTAGE =					0x40
};

/**
 * @brief SHT11 state data
 *
 * Stores importend runtime data,
 */
static struct _SHT11_data {
	bool sht11;									///< true if SHT11 present, else false
	bool busy;									///< true if SHT11 busy (already does a measurement), else false
	uint8_t crc;									///< CRC8 value of all send and received bytes
	SHT11_resolution resolution;		///< conversion resolution
} SHT11_data;


static uint8_t const crc8_table[256] = {
0,  49 , 98, 83, 196, 245, 166, 151, 185, 136, 219, 234, 125, 76, 31, 46, 67, 114, 33, 16, 135, 182, 229, 212, 250, 203, 152, 169, 62, 15, 92, 109,
134, 183, 228, 213, 66, 115, 32, 17, 63, 14, 93, 108, 251, 202, 153, 168, 197, 244, 167, 150, 1, 48,  99,  82, 124, 77, 30, 47, 184, 137, 218, 235,
61, 12, 95, 110, 249, 200, 155, 170, 132, 181, 230, 215, 64, 113, 34, 19, 126, 79, 28, 45, 186, 139, 216, 233, 199, 246, 165, 148, 3, 50, 97, 80,
187, 138, 217, 232, 127, 78, 29, 44, 2, 51, 96, 81, 198, 247, 164, 149, 248, 201, 154, 171, 60, 13, 94, 111, 65, 112, 35, 18, 133, 180, 231, 214,
122,  75,  24,  41, 190, 143, 220, 237, 195, 242, 161, 144, 7, 54, 101, 84, 57, 8, 91, 106, 253, 204, 159, 174, 128, 177, 226, 211, 68, 117, 38, 23,
252, 205, 158, 175,  56, 9, 90, 107, 69, 116, 39, 22, 129, 176, 227, 210, 191, 142, 221, 236, 123, 74, 25, 40, 6, 55, 100, 85, 194, 243, 160, 145,
71, 118, 37, 20, 131, 178, 225, 208, 254, 207, 156, 173, 58, 11, 88, 105, 4, 53, 102, 87, 192, 241, 162, 147, 189, 140, 223, 238, 121, 72, 27, 42,
193, 240, 163, 146, 5, 52, 103, 86, 120, 73, 26, 43, 188, 141, 222, 239, 130, 179, 224, 209, 70, 119, 36, 21, 59, 10, 89, 104, 255, 206, 157, 172
};

/**
 * @brief Update CRC8
 *
 * Update the stored CRC8 value by adding a new character.
 * @param b add this character to the checksum
 */
void SHT11_updateCRC8(uint8_t b);

/**
 * @brief SHT11 crc check
 *
 * Check the returned 8 bit crc value against the
 * computed one.
 * @return true on error, else false
 */
bool SHT11_checkCRC(uint8_t c);

/**
 * @name SHT11 physical layer access functions
 * @see sd_state_t
 * @{
 */
inline void SHT11_CLK_LOW()		{ P3OUT &= ~0x20; }
inline void SHT11_CLK_HIGH()		{ P3OUT |=  0x20; }

inline void SHT11_DATA_LOW()	{ P3OUT &= ~0x10; }
inline void SHT11_DATA_HIGH()	{ P3OUT |=  0x10; }
inline uint16_t SHT11_DATA()		{ return P3IN & 0x10; }

inline void SHT11_OUTPUT()			{ P3DIR |= 0x10; }
inline void SHT11_INPUT()			{ P3DIR &= ~0x10; }
/** @} */

/**
 * @brief write to Sensibus
 *
 * Writes a byte on the Sensibus and checks the acknowledgement.
 * @param value byte to write
 * @return true on error, else false
 */
bool SHT11_write_byte(uint8_t value);

/**
 * @brief read from Sensibus
 *
 * Reads a byte from the Sensibus and gives an acknowledgement in case of "ack=1"
 *@param crcUpdate set to true to automatically update the crc value, else false
 * @return read value
 */
uint8_t SHT11_read_byte(bool crcUpdate);

/**
 * @brief Transmission start sequence
 *
 *  Generates a transmission start sequence.
 */
void SHT11_transstart();

/**
 * @brief Hardware reset
 *
 * Communication reset: DATA-line=1 and at least 9 SCK cycles followed by transstart.
 * \code
 *        _____________________________________________________         _____
 *  DATA:                                                      |_______|
 *           _    _    _    _    _    _    _    _    _        ___     ___
 *  SCK : __| |__| |__| |__| |__| |__| |__| |__| |__| |______|   |___|   |___
 * \endcode
 * @see SHT11_softreset
 */
void SHT11_connectionreset();

/**
 * @brief Software reset
 *
 * Send reset command to SHT11.
 * @see SHT11_connectionreset
 */
bool SHT11_softreset();

/**
 * @brief Read status register
 *
 * Reads the status register
 * @param p_value destination of status register content
 * @return true on error, else false
 */
bool SHT11_read_statusreg(uint8_t *p_value);

/**
 * @brief Write status register
 *
 * Writes the status register.
 * @param value value to write to the status register
 * @return true on error, else false
 */
bool SHT11_write_statusreg(uint8_t p_value);

/**
 * @brief Read value from SHT11
 *
 * Does a measurement (humidity or temperature).
 * @param p_value destination of result
 * @param command  MEASURE_TEMP or MEASURE_HUMI
 * @return true on error, else false
 */
bool SHT11_measure(uint16_t *p_value, SHT11_command command);

/**
 * @brief Calculate temperature from raw value
 *
 * Calculates temperature in C or K from the raw read value.
 * @param rawTemperature raw temperature value
 * @param unit output unit
 *	@return temperature in C or K
 */
float SHT11_calcTemp(uint16_t rawTemperature, SHT11_tempUnit unit);

/**
 * @brief Calculate (uncorrected) humidity from raw value
 *
 * Calculates relative humidity from the raw read value without
 * temperature compensation.
 * @param rawHumidity raw humidity value
 *	@return humidity in %RH
 * @see SHT11_calcHumiCorrected
 */
float SHT11_calcHumi(uint16_t rawHumidity);

/**
 * @brief Calculate (corrected) humidity from raw value
 *
 * Calculates relative humidity from the raw read value.
 * @param rawHumidity raw temperature value
 * @param humidity uncorrected humidity in %RH
 * @param temperature temperature in C
 *	@return humidity in %RH
 * @see SHT11_calcHumi
 */
float SHT11_calcHumiCorrected(uint16_t rawHumidity, float humidity, float temperature);

/**
 * @brief check if SHT11 present
 *
 * @return true if present, else false
 */
bool SHT11_checkForSHT11();

#endif /*SHT11_INT_H_*/
